import sqlite3, socket, xmlrpclib, time, os
import utils


class PeerManager:

    # internal ip = temp arg/attribute
    def __init__(self, db_dir_path, server, server_port, refresh_time, address,
                       internal_ip, prefix, manual, pp , db_size):
        self._refresh_time = refresh_time
        self._address = address
        self._internal_ip = internal_ip
        self._prefix = prefix
        self._server = server
        self._server_port = server_port
        self._db_size = db_size
        self._pp = pp
        self._blacklist = [(prefix,)]
        self._manual = manual

        self._proxy = xmlrpclib.ServerProxy('http://%s:%u'
                % (server, server_port))

        utils.log('Connectiong to peers database...', 4)
        self._db = sqlite3.connect(os.path.join(db_dir_path, 'peers.db'),
                                   isolation_level=None)
        utils.log('Database opened', 5)
        utils.log('Preparing peers database...', 4)
        try:
            self._db.execute("UPDATE peers SET used = 0")
            self._db.execute("""CREATE TABLE IF NOT EXISTS blacklist (
                                prefix TEXT PRIMARY KEY,
                                flag INTEGER NOT NULL)""")
            self._db.execute("""CREATE INDEX IF NOT EXISTS
                                blacklist_flag ON blacklist(flag)""")
            utils.log('Database prepared', 5)
        except sqlite3.OperationalError, e:
            if e.args[0] == 'no such table: peers':
                raise RuntimeError
            else:
                utils.log(e, 1)
        except sqlite3.Error, e:
            utils.log(e, 1)

        self.next_refresh = time.time()

    def clear_blacklist(self, flag):
        utils.log('Clearing blacklist from flag %u' % (flag,), 3)
        try:
            with self._db:
                self._db.execute("DELETE FROM blacklist WHERE flag = ?",
                                  (flag,))
                utils.log('Blacklist cleared', 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def blacklist(self, prefix, flag):
        utils.log('Blacklisting %s' % (prefix,), 4)
        try:
            with self._db:
                self._db.execute("DELETE FROM peers WHERE prefix = ?",
                                (prefix,))
                self._db.execute("""INSERT OR REPLACE INTO blacklist
                                    VALUES (?,?)""", (prefix, flag))
                utils.log('%s blacklisted' % (prefix,), 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def whitelist(self, prefix):
        utils.log('Unblacklisting %s' % (prefix,), 4)
        try:
            with self._db:
                self._db.execute("DELETE FROM blacklist WHERE prefix = ?", (prefix,))
                utils.log('%s whitelisted' % (prefix,), 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def refresh(self):
        utils.log('Refreshing the peers DB...', 2)
        try:
            self._declare()
            self._populate()
            utils.log('DB refreshed', 3)
            self.next_refresh = time.time() + self._refresh_time
        except socket.error, e:
            utils.log(e, 4)
            utils.log('Connection to server failed, retrying in 30s', 2)
            self.next_refresh = time.time() + 30

    def _declare(self):
        if self._address != None:
            utils.log('Sending connection info to server...', 3)
            self._proxy.declare((self._internal_ip,
                    utils.address_str(self._address)))
            utils.log('Info sent', 5)
        else:
            utils.log("Warning : couldn't send ip, unknown external config", 4)

    def _populate(self):
        utils.log('Populating the peers DB...', 2)
        new_peer_list = self._proxy.getPeerList(self._db_size,
                self._internal_ip)
        try:
            with self._db:
                self._db.execute("""DELETE FROM peers WHERE used <= 0 ORDER BY
                                    used,random() LIMIT MAX(0, ? + (SELECT
                                    COUNT(*) FROM peers WHERE used <= 0))""",
                                    (str(len(new_peer_list) - self._db_size),))
                self._db.executemany("""INSERT OR IGNORE INTO peers (prefix, address)
                                        VALUES (?,?)""", new_peer_list)
                self._db.execute("""DELETE FROM peers WHERE prefix IN
                                    (SELECT prefix FROM blacklist)""")
                utils.log('DB populated', 3)
                utils.log('New peers : %s' % ', '.join(map(str, new_peer_list)), 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def getUnusedPeers(self, peer_count):
        try:
            return self._db.execute("""SELECT prefix, address FROM peers WHERE used
                                       <= 0 ORDER BY used DESC,RANDOM() LIMIT ?""",
                                       (peer_count,))
        except sqlite3.Error, e:
            utils.log(e, 1)
            return [('0', '')]

    def usePeer(self, prefix):
        utils.log('Updating peers database : using peer ' + str(prefix), 5)
        try:
            with self._db:
                self._db.execute("UPDATE peers SET used = 1 WHERE prefix = ?",
                        (prefix,))
                utils.log('DB updated', 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def unusePeer(self, prefix):
        utils.log('Updating peers database : unusing peer ' + str(prefix), 5)
        try:
            with self._db:
                self._db.execute("UPDATE peers SET used = 0 WHERE prefix = ?",
                        (prefix,))
                utils.log('DB updated', 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def flagPeer(self, prefix):
        utils.log('Updating peers database : flagging peer ' + str(prefix), 5)
        try:
            with self._db:
                self._db.execute("UPDATE peers SET used = -1 WHERE prefix = ?",
                         (prefix,))
                utils.log('DB updated', 5)
        except sqlite3.Error, e:
            utils.log(e, 1)

    def handle_message(self, msg):
        script_type, arg = msg.split()
        if script_type == 'client-connect':
            utils.log('Incomming connection from %s' % (arg,), 3)
        elif script_type == 'client-disconnect':
            utils.log('%s has disconnected' % (arg,), 3)
        elif script_type == 'route-up':
            if not self._manual:
                external_ip = arg
                new_address = list([external_ip, port, proto]
                                   for port, proto in self._pp)
                if self._address != new_address:
                    self._address = new_address
                    utils.log('Received new external ip : %s' 
                              % (external_ip,), 3)
                    self._declare()
        else:
            utils.log('Unknow message recieved from the openvpn pipe : '
                    + msg, 1)

